import { Construct } from "constructs";
import { ENVIRONMENT_PLACEHOLDER, getParameterPath, getParameterFallbackPath } from "./util/functionBinding.js";
/**
 * The `Secret` construct is a higher level CDK construct that makes it easy to manage app secrets.
 *
 * @example
 * ### Using the minimal config
 *
 * ```js
 * import { Config } from "@serverless-stack/resources";
 *
 * new Config.Secret(stack, "STRIPE_KEY");
 * ```
 */
export class Secret extends Construct {
    constructor(scope, id) {
        super(scope, id);
        this.id = id;
        this.name = id;
    }
    /** @internal */
    getConstructMetadata() {
        return {
            type: "Secret",
            data: {
                name: this.name,
            },
        };
    }
    /** @internal */
    getFunctionBinding() {
        const app = this.node.root;
        return {
            clientPackage: "config",
            variables: {
                value: {
                    environment: ENVIRONMENT_PLACEHOLDER,
                    // SSM parameters will be created manually via CLI
                    parameter: undefined,
                },
            },
            permissions: {
                "ssm:GetParameters": [
                    `arn:aws:ssm:${app.region}:${app.account}:parameter${getParameterPath(this, "value")}`,
                    `arn:aws:ssm:${app.region}:${app.account}:parameter${getParameterFallbackPath(this, "value")}`
                ],
            },
        };
    }
    static create(scope, ...parameters) {
        const result = {};
        for (const name of parameters) {
            result[name] = new Secret(scope, name);
        }
        return result;
    }
}
